@@ -6,11 +6,17 @@ import traceback
 import types
 import copy
 import operator
+import random
+import string
 from memoize import Memoize
 
 
 class _Symbolic(tuple):
 
+  def match(self, other, valuestore=None):
+    import match
+    return match.match(self, other, valuestore)
+
   def walk(self, *fns):
     if len(fns) > 1:
       def _(exp):
@@ -46,13 +52,6 @@ class _Symbolic(tuple):
         'id': id(self)
         }
 
-  def _canonicalize(self):
-    '''
-    overridden by some subtypes
-     - should return a canonical version of itself
-    '''
-    return self
-
   def substitute(self, subs):
     '''
     takes a dictionary of substitutions
@@ -145,32 +144,51 @@ class _Symbolic(tuple):
   def __neg__(self):
     return self * -1
 
-class Boolean(int):
+class _KnownValue(_Symbolic):
+  def value(self):
+    raise BaseException('not implemented')
 
+class Boolean(_KnownValue):
+
+  @Memoize
   def __new__(typ, b):
-    self = int.__new__(typ, 1 if b else 0)
+    self = _KnownValue.__new__(typ)
     self.name = str(b)
     self.boolean = b
     return self
 
+  def value(self):
+    return bool(self.boolean)
+
   def __str__(self):
     return str(self.boolean)
 
   def __repr__(self):
     return str(self)
 
-class Number(_Symbolic):
+  def __eq__(self, other):
+    if isinstance(other, Boolean):
+      return bool(self.boolean) == bool(other.boolean)
+    elif isinstance(other, _Symbolic):
+      return other.__eq__(self)
+    else:
+      return bool(self.boolean) == other
+
+class Number(_KnownValue):
 
   IFORMAT = str
   FFORMAT = str
 
   def __new__(typ, n):
     n = float(n)
-    self = _Symbolic.__new__(typ)
+    self = _KnownValue.__new__(typ)
     self.name = str(n)
     self.n = n
     return self
 
+  def value(self):
+    return self.n
+
   def __eq__(self, other):
     if isinstance(other, _Symbolic):
       return super(Number, self).__eq__(other)
@@ -194,18 +212,46 @@ class Number(_Symbolic):
 
 class Wild(_Symbolic):
   '''
-  wilds will not be equal even if they have the same name
-  but the same *instance* will be equal to itself
-
-  the main part of this is for substituting patterns -
-   this is not implemented yet
+  wilds will be equal to anything, and are used for pattern matching
   '''
 
   def __new__(typ, name, **kargs):
     self = _Symbolic.__new__(typ)
     self.name = name
     self.kargs = kargs
-    self.iswild = True
+    return self
+
+  def __eq__(self, other):
+    return other.name == self.name if type(other) == type(self) else False
+
+  def __str__(self):
+    return self.name
+
+  def __repr__(self):
+    return str(self)
+
+  def __call__(self, *args):
+    return Fn(self, *args)
+
+  def _dump(self):
+    return {
+        'type': type(self),
+        'name': self.name,
+        'kargs': self.kargs,
+        'id': id(self)
+        }
+
+class Symbol(_Symbolic):
+  '''
+  symbols with the same name and kargs will be equal
+  (and in fact are guaranteed to be the same instance)
+  '''
+
+  @Memoize
+  def __new__(typ, name, **kargs):
+    self = Wild.__new__(typ, name)
+    self.name = name
+    self.kargs = kargs
     return self
 
   def __eq__(self, other):
@@ -225,23 +271,9 @@ class Wild(_Symbolic):
         'type': type(self),
         'name': self.name,
         'kargs': self.kargs,
-        'iswild': self.iswild,
         'id': id(self)
         }
 
-class Symbol(Wild):
-  '''
-  symbols with the same name will be equal
-  (and in fact are wilds guaranteed to be the same instance)
-  '''
-
-  @Memoize
-  def __new__(typ, name, **kargs):
-    self = Wild.__new__(typ, name)
-    self.name = name
-    self.kargs = kargs
-    self.iswild = False
-    return self
 
 class Fn(_Symbolic):
 
@@ -265,28 +297,12 @@ class Fn(_Symbolic):
     kargs = fn.kargs
     self.kargs = fn.kargs
 
-    if len(args) == 2 and 'numeric' in kargs:
-      x = args[0]
-      y = args[1]
-      if isinstance(x, Number) and isinstance(y, Number):
-        if 'cast' in kargs and kargs['cast'] != None:
-          x = kargs['cast'](x.n)
-          y = kargs['cast'](y.n)
-        else:
-          x = x.n
-          y = y.n
-        try:
-          nfn = getattr(operator, kargs['numeric'])
-          return symbolic(nfn(x,y))
-        except:
-          raise BaseException("Could not %s %s %s" % (x, kargs['numeric'], y))
-
     self.name = fn.name
     self.fn = fn
     self.args = args
 
     import simplify
-    rv = simplify.simplify(self._canonicalize())._canonicalize()
+    rv = simplify.simplify(self)
 
     return rv
 
@@ -361,25 +377,6 @@ class Fn(_Symbolic):
 
     return rv
 
-  def _canonicalize(self):
-    import simplify
-    # canonicalize the arguments first
-    args = list(map(lambda x: x._canonicalize(), self.args))
-    if tuple(args) != tuple(self.args):
-      self = Fn(self.fn, *args)
-
-    # if it's associative and one of the arguments is another instance of the
-    # same function, canonicalize the order
-    if len(self.args) == 2 and 'associative' in self.kargs and self.kargs['associative']:
-      args = self._get_assoc_arguments()
-      oldargs = tuple(args)
-      args.sort(simplify._order)
-      if tuple(args) != oldargs:
-        kargs = copy.copy(self.kargs)
-        self = reduce(lambda a, b: Fn(self.fn, a, b), args)
-
-    return self
-
   @staticmethod
   def LessThan(lhs, rhs):
     return Fn(stdops.LessThan, lhs, rhs)
@@ -460,21 +457,25 @@ def symbols(symstr, **kargs):
 
   return tuple(rv)
 
-def wilds(symstr):
+def wilds(symstr, **kargs):
   '''
-  takes a string of variable names seperated by whitespace
-  returns a tuple of wilds
+  wilds should match anything
   '''
   syms = symstr.split(' ')
   if len(syms) == 1:
-    return Wild(syms[0])
+    return Wild(syms[0], **kargs)
 
   rv = []
   for i in syms:
-    rv.append(Wild(i))
+    rv.append(Wild(i, **kargs))
 
   return tuple(rv)
 
+def wild(name=None, **kargs):
+  if name == None:
+    name = ''.join(random.choice(string.ascii_lowercase) for x in range(12))
+  return Wild(name, **kargs)
+
 def symbolic(obj, **kargs): 
   '''
   makes the symbolic version of an object
